pragma Warnings (Off);
pragma Ada_95;
pragma Source_File_Name (ada_main, Spec_File_Name => "b_gnatm.ads");
pragma Source_File_Name (ada_main, Body_File_Name => "b_gnatm.adb");
pragma Suppress (Overflow_Check);
with Ada.Exceptions;

package body ada_main is

   E079 : Short_Integer; pragma Import (Ada, E079, "system__os_lib_E");
   E011 : Short_Integer; pragma Import (Ada, E011, "ada__exceptions_E");
   E016 : Short_Integer; pragma Import (Ada, E016, "system__soft_links_E");
   E026 : Short_Integer; pragma Import (Ada, E026, "system__exception_table_E");
   E049 : Short_Integer; pragma Import (Ada, E049, "ada__containers_E");
   E075 : Short_Integer; pragma Import (Ada, E075, "ada__io_exceptions_E");
   E033 : Short_Integer; pragma Import (Ada, E033, "ada__numerics_E");
   E061 : Short_Integer; pragma Import (Ada, E061, "ada__strings_E");
   E063 : Short_Integer; pragma Import (Ada, E063, "ada__strings__maps_E");
   E066 : Short_Integer; pragma Import (Ada, E066, "ada__strings__maps__constants_E");
   E086 : Short_Integer; pragma Import (Ada, E086, "interfaces__c_E");
   E027 : Short_Integer; pragma Import (Ada, E027, "system__exceptions_E");
   E090 : Short_Integer; pragma Import (Ada, E090, "system__object_reader_E");
   E056 : Short_Integer; pragma Import (Ada, E056, "system__dwarf_lines_E");
   E018 : Short_Integer; pragma Import (Ada, E018, "system__soft_links__initialize_E");
   E048 : Short_Integer; pragma Import (Ada, E048, "system__traceback__symbolic_E");
   E032 : Short_Integer; pragma Import (Ada, E032, "system__img_int_E");
   E070 : Short_Integer; pragma Import (Ada, E070, "system__img_uns_E");
   E198 : Short_Integer; pragma Import (Ada, E198, "ada__assertions_E");
   E130 : Short_Integer; pragma Import (Ada, E130, "ada__strings__utf_encoding_E");
   E138 : Short_Integer; pragma Import (Ada, E138, "ada__tags_E");
   E128 : Short_Integer; pragma Import (Ada, E128, "ada__strings__text_buffers_E");
   E190 : Short_Integer; pragma Import (Ada, E190, "gnat_E");
   E126 : Short_Integer; pragma Import (Ada, E126, "ada__streams_E");
   E179 : Short_Integer; pragma Import (Ada, E179, "system__file_control_block_E");
   E148 : Short_Integer; pragma Import (Ada, E148, "system__finalization_root_E");
   E124 : Short_Integer; pragma Import (Ada, E124, "ada__finalization_E");
   E176 : Short_Integer; pragma Import (Ada, E176, "system__file_io_E");
   E183 : Short_Integer; pragma Import (Ada, E183, "system__storage_pools_E");
   E181 : Short_Integer; pragma Import (Ada, E181, "system__finalization_masters_E");
   E315 : Short_Integer; pragma Import (Ada, E315, "system__storage_pools__subpools_E");
   E160 : Short_Integer; pragma Import (Ada, E160, "ada__strings__unbounded_E");
   E111 : Short_Integer; pragma Import (Ada, E111, "ada__calendar_E");
   E308 : Short_Integer; pragma Import (Ada, E308, "ada__text_io_E");
   E265 : Short_Integer; pragma Import (Ada, E265, "gnat__byte_order_mark_E");
   E321 : Short_Integer; pragma Import (Ada, E321, "gnat__ctrl_c_E");
   E310 : Short_Integer; pragma Import (Ada, E310, "gnat__directory_operations_E");
   E220 : Short_Integer; pragma Import (Ada, E220, "gnat__dynamic_htables_E");
   E313 : Short_Integer; pragma Import (Ada, E313, "system__pool_global_E");
   E185 : Short_Integer; pragma Import (Ada, E185, "system__regexp_E");
   E109 : Short_Integer; pragma Import (Ada, E109, "ada__directories_E");
   E306 : Short_Integer; pragma Import (Ada, E306, "gnat__command_line_E");
   E203 : Short_Integer; pragma Import (Ada, E203, "types_E");
   E006 : Short_Integer; pragma Import (Ada, E006, "gnatvsn_E");
   E199 : Short_Integer; pragma Import (Ada, E199, "hostparm_E");
   E192 : Short_Integer; pragma Import (Ada, E192, "opt_E");
   E205 : Short_Integer; pragma Import (Ada, E205, "output_E");
   E304 : Short_Integer; pragma Import (Ada, E304, "sfn_scan_E");
   E286 : Short_Integer; pragma Import (Ada, E286, "stylesw_E");
   E253 : Short_Integer; pragma Import (Ada, E253, "elists_E");
   E245 : Short_Integer; pragma Import (Ada, E245, "uintp_E");
   E247 : Short_Integer; pragma Import (Ada, E247, "urealp_E");
   E278 : Short_Integer; pragma Import (Ada, E278, "warnsw_E");
   E211 : Short_Integer; pragma Import (Ada, E211, "widechar_E");
   E207 : Short_Integer; pragma Import (Ada, E207, "namet_E");
   E189 : Short_Integer; pragma Import (Ada, E189, "butil_E");
   E269 : Short_Integer; pragma Import (Ada, E269, "binderr_E");
   E235 : Short_Integer; pragma Import (Ada, E235, "casing_E");
   E270 : Short_Integer; pragma Import (Ada, E270, "err_vars_E");
   E224 : Short_Integer; pragma Import (Ada, E224, "fmap_E");
   E222 : Short_Integer; pragma Import (Ada, E222, "osint_E");
   E228 : Short_Integer; pragma Import (Ada, E228, "sdefault_E");
   E230 : Short_Integer; pragma Import (Ada, E230, "targparm_E");
   E218 : Short_Integer; pragma Import (Ada, E218, "fname_E");
   E334 : Short_Integer; pragma Import (Ada, E334, "osint__m_E");
   E233 : Short_Integer; pragma Import (Ada, E233, "snames_E");
   E187 : Short_Integer; pragma Import (Ada, E187, "ali_E");
   E267 : Short_Integer; pragma Import (Ada, E267, "scans_E");
   E241 : Short_Integer; pragma Import (Ada, E241, "aspects_E");
   E255 : Short_Integer; pragma Import (Ada, E255, "nlists_E");
   E258 : Short_Integer; pragma Import (Ada, E258, "sinfo__nodes_E");
   E249 : Short_Integer; pragma Import (Ada, E249, "einfo__entities_E");
   E251 : Short_Integer; pragma Import (Ada, E251, "einfo__utils_E");
   E239 : Short_Integer; pragma Import (Ada, E239, "atree_E");
   E260 : Short_Integer; pragma Import (Ada, E260, "sinfo__utils_E");
   E263 : Short_Integer; pragma Import (Ada, E263, "sinput_E");
   E284 : Short_Integer; pragma Import (Ada, E284, "sem_aux_E");
   E296 : Short_Integer; pragma Import (Ada, E296, "sinput__c_E");
   E276 : Short_Integer; pragma Import (Ada, E276, "stringt_E");
   E274 : Short_Integer; pragma Import (Ada, E274, "erroutc_E");
   E280 : Short_Integer; pragma Import (Ada, E280, "lib_E");
   E282 : Short_Integer; pragma Import (Ada, E282, "uname_E");
   E272 : Short_Integer; pragma Import (Ada, E272, "errout_E");
   E300 : Short_Integer; pragma Import (Ada, E300, "fname__uf_E");
   E298 : Short_Integer; pragma Import (Ada, E298, "fname__sf_E");
   E294 : Short_Integer; pragma Import (Ada, E294, "styleg_E");
   E325 : Short_Integer; pragma Import (Ada, E325, "errutil_E");
   E323 : Short_Integer; pragma Import (Ada, E323, "make_util_E");
   E288 : Short_Integer; pragma Import (Ada, E288, "scng_E");
   E237 : Short_Integer; pragma Import (Ada, E237, "ali__util_E");
   E330 : Short_Integer; pragma Import (Ada, E330, "switch_E");
   E336 : Short_Integer; pragma Import (Ada, E336, "switch__m_E");
   E340 : Short_Integer; pragma Import (Ada, E340, "tempdir_E");
   E332 : Short_Integer; pragma Import (Ada, E332, "usage_E");
   E328 : Short_Integer; pragma Import (Ada, E328, "makeusg_E");
   E105 : Short_Integer; pragma Import (Ada, E105, "make_E");
   E002 : Short_Integer; pragma Import (Ada, E002, "gnatmake_E");

   Sec_Default_Sized_Stacks : array (1 .. 1) of aliased System.Secondary_Stack.SS_Stack (System.Parameters.Runtime_Default_Sec_Stack_Size);

   Local_Priority_Specific_Dispatching : constant String := "";
   Local_Interrupt_States : constant String := "";

   Is_Elaborated : Boolean := False;

   procedure finalize_library is
   begin
      declare
         procedure F1;
         pragma Import (Ada, F1, "ada__directories__finalize_body");
      begin
         E109 := E109 - 1;
         F1;
      end;
      declare
         procedure F2;
         pragma Import (Ada, F2, "ada__directories__finalize_spec");
      begin
         F2;
      end;
      E185 := E185 - 1;
      declare
         procedure F3;
         pragma Import (Ada, F3, "system__regexp__finalize_spec");
      begin
         F3;
      end;
      E313 := E313 - 1;
      declare
         procedure F4;
         pragma Import (Ada, F4, "system__pool_global__finalize_spec");
      begin
         F4;
      end;
      E308 := E308 - 1;
      declare
         procedure F5;
         pragma Import (Ada, F5, "ada__text_io__finalize_spec");
      begin
         F5;
      end;
      E160 := E160 - 1;
      declare
         procedure F6;
         pragma Import (Ada, F6, "ada__strings__unbounded__finalize_spec");
      begin
         F6;
      end;
      E315 := E315 - 1;
      declare
         procedure F7;
         pragma Import (Ada, F7, "system__storage_pools__subpools__finalize_spec");
      begin
         F7;
      end;
      E181 := E181 - 1;
      declare
         procedure F8;
         pragma Import (Ada, F8, "system__finalization_masters__finalize_spec");
      begin
         F8;
      end;
      declare
         procedure F9;
         pragma Import (Ada, F9, "system__file_io__finalize_body");
      begin
         E176 := E176 - 1;
         F9;
      end;
      declare
         procedure Reraise_Library_Exception_If_Any;
            pragma Import (Ada, Reraise_Library_Exception_If_Any, "__gnat_reraise_library_exception_if_any");
      begin
         Reraise_Library_Exception_If_Any;
      end;
   end finalize_library;

   procedure adafinal is
      procedure s_stalib_adafinal;
      pragma Import (Ada, s_stalib_adafinal, "system__standard_library__adafinal");

      procedure Runtime_Finalize;
      pragma Import (C, Runtime_Finalize, "__gnat_runtime_finalize");

   begin
      if not Is_Elaborated then
         return;
      end if;
      Is_Elaborated := False;
      Runtime_Finalize;
      s_stalib_adafinal;
   end adafinal;

   type No_Param_Proc is access procedure;
   pragma Favor_Top_Level (No_Param_Proc);

   procedure adainit is
      Main_Priority : Integer;
      pragma Import (C, Main_Priority, "__gl_main_priority");
      Time_Slice_Value : Integer;
      pragma Import (C, Time_Slice_Value, "__gl_time_slice_val");
      WC_Encoding : Character;
      pragma Import (C, WC_Encoding, "__gl_wc_encoding");
      Locking_Policy : Character;
      pragma Import (C, Locking_Policy, "__gl_locking_policy");
      Queuing_Policy : Character;
      pragma Import (C, Queuing_Policy, "__gl_queuing_policy");
      Task_Dispatching_Policy : Character;
      pragma Import (C, Task_Dispatching_Policy, "__gl_task_dispatching_policy");
      Priority_Specific_Dispatching : System.Address;
      pragma Import (C, Priority_Specific_Dispatching, "__gl_priority_specific_dispatching");
      Num_Specific_Dispatching : Integer;
      pragma Import (C, Num_Specific_Dispatching, "__gl_num_specific_dispatching");
      Main_CPU : Integer;
      pragma Import (C, Main_CPU, "__gl_main_cpu");
      Interrupt_States : System.Address;
      pragma Import (C, Interrupt_States, "__gl_interrupt_states");
      Num_Interrupt_States : Integer;
      pragma Import (C, Num_Interrupt_States, "__gl_num_interrupt_states");
      Unreserve_All_Interrupts : Integer;
      pragma Import (C, Unreserve_All_Interrupts, "__gl_unreserve_all_interrupts");
      Detect_Blocking : Integer;
      pragma Import (C, Detect_Blocking, "__gl_detect_blocking");
      Default_Stack_Size : Integer;
      pragma Import (C, Default_Stack_Size, "__gl_default_stack_size");
      Default_Secondary_Stack_Size : System.Parameters.Size_Type;
      pragma Import (C, Default_Secondary_Stack_Size, "__gnat_default_ss_size");
      Bind_Env_Addr : System.Address;
      pragma Import (C, Bind_Env_Addr, "__gl_bind_env_addr");

      procedure Runtime_Initialize (Install_Handler : Integer);
      pragma Import (C, Runtime_Initialize, "__gnat_runtime_initialize");

      Finalize_Library_Objects : No_Param_Proc;
      pragma Import (C, Finalize_Library_Objects, "__gnat_finalize_library_objects");
      Binder_Sec_Stacks_Count : Natural;
      pragma Import (Ada, Binder_Sec_Stacks_Count, "__gnat_binder_ss_count");
      Default_Sized_SS_Pool : System.Address;
      pragma Import (Ada, Default_Sized_SS_Pool, "__gnat_default_ss_pool");

   begin
      if Is_Elaborated then
         return;
      end if;
      Is_Elaborated := True;
      Main_Priority := -1;
      Time_Slice_Value := -1;
      WC_Encoding := 'b';
      Locking_Policy := ' ';
      Queuing_Policy := ' ';
      Task_Dispatching_Policy := ' ';
      Priority_Specific_Dispatching :=
        Local_Priority_Specific_Dispatching'Address;
      Num_Specific_Dispatching := 0;
      Main_CPU := -1;
      Interrupt_States := Local_Interrupt_States'Address;
      Num_Interrupt_States := 0;
      Unreserve_All_Interrupts := 0;
      Detect_Blocking := 0;
      Default_Stack_Size := -1;

      ada_main'Elab_Body;
      Default_Secondary_Stack_Size := System.Parameters.Runtime_Default_Sec_Stack_Size;
      Binder_Sec_Stacks_Count := 1;
      Default_Sized_SS_Pool := Sec_Default_Sized_Stacks'Address;

      Runtime_Initialize (1);

      Finalize_Library_Objects := finalize_library'access;

      Ada.Exceptions'Elab_Spec;
      System.Soft_Links'Elab_Spec;
      System.Exception_Table'Elab_Body;
      E026 := E026 + 1;
      Ada.Containers'Elab_Spec;
      E049 := E049 + 1;
      Ada.Io_Exceptions'Elab_Spec;
      E075 := E075 + 1;
      Ada.Numerics'Elab_Spec;
      E033 := E033 + 1;
      Ada.Strings'Elab_Spec;
      E061 := E061 + 1;
      Ada.Strings.Maps'Elab_Spec;
      E063 := E063 + 1;
      Ada.Strings.Maps.Constants'Elab_Spec;
      E066 := E066 + 1;
      Interfaces.C'Elab_Spec;
      E086 := E086 + 1;
      System.Exceptions'Elab_Spec;
      E027 := E027 + 1;
      System.Object_Reader'Elab_Spec;
      E090 := E090 + 1;
      System.Dwarf_Lines'Elab_Spec;
      System.Os_Lib'Elab_Body;
      E079 := E079 + 1;
      System.Soft_Links.Initialize'Elab_Body;
      E018 := E018 + 1;
      E016 := E016 + 1;
      System.Traceback.Symbolic'Elab_Body;
      E048 := E048 + 1;
      System.Img_Int'Elab_Spec;
      E032 := E032 + 1;
      E011 := E011 + 1;
      System.Img_Uns'Elab_Spec;
      E070 := E070 + 1;
      E056 := E056 + 1;
      Ada.Assertions'Elab_Spec;
      E198 := E198 + 1;
      Ada.Strings.Utf_Encoding'Elab_Spec;
      E130 := E130 + 1;
      Ada.Tags'Elab_Spec;
      Ada.Tags'Elab_Body;
      E138 := E138 + 1;
      Ada.Strings.Text_Buffers'Elab_Spec;
      E128 := E128 + 1;
      Gnat'Elab_Spec;
      E190 := E190 + 1;
      Ada.Streams'Elab_Spec;
      E126 := E126 + 1;
      System.File_Control_Block'Elab_Spec;
      E179 := E179 + 1;
      System.Finalization_Root'Elab_Spec;
      E148 := E148 + 1;
      Ada.Finalization'Elab_Spec;
      E124 := E124 + 1;
      System.File_Io'Elab_Body;
      E176 := E176 + 1;
      System.Storage_Pools'Elab_Spec;
      E183 := E183 + 1;
      System.Finalization_Masters'Elab_Spec;
      System.Finalization_Masters'Elab_Body;
      E181 := E181 + 1;
      System.Storage_Pools.Subpools'Elab_Spec;
      E315 := E315 + 1;
      Ada.Strings.Unbounded'Elab_Spec;
      E160 := E160 + 1;
      Ada.Calendar'Elab_Spec;
      Ada.Calendar'Elab_Body;
      E111 := E111 + 1;
      Ada.Text_Io'Elab_Spec;
      Ada.Text_Io'Elab_Body;
      E308 := E308 + 1;
      E265 := E265 + 1;
      E321 := E321 + 1;
      Gnat.Directory_Operations'Elab_Spec;
      Gnat.Directory_Operations'Elab_Body;
      E310 := E310 + 1;
      E220 := E220 + 1;
      System.Pool_Global'Elab_Spec;
      E313 := E313 + 1;
      System.Regexp'Elab_Spec;
      E185 := E185 + 1;
      Ada.Directories'Elab_Spec;
      Ada.Directories'Elab_Body;
      E109 := E109 + 1;
      Gnat.Command_Line'Elab_Spec;
      Gnat.Command_Line'Elab_Body;
      E306 := E306 + 1;
      Types'Elab_Spec;
      E203 := E203 + 1;
      E006 := E006 + 1;
      Hostparm'Elab_Spec;
      E199 := E199 + 1;
      Opt'Elab_Spec;
      E192 := E192 + 1;
      Output'Elab_Body;
      E205 := E205 + 1;
      Sfn_Scan'Elab_Spec;
      E304 := E304 + 1;
      E286 := E286 + 1;
      Elists'Elab_Body;
      E253 := E253 + 1;
      Uintp'Elab_Spec;
      Uintp'Elab_Body;
      E245 := E245 + 1;
      Urealp'Elab_Body;
      E247 := E247 + 1;
      E278 := E278 + 1;
      E211 := E211 + 1;
      Namet'Elab_Spec;
      Namet'Elab_Body;
      E207 := E207 + 1;
      E189 := E189 + 1;
      E269 := E269 + 1;
      E235 := E235 + 1;
      Err_Vars'Elab_Spec;
      E270 := E270 + 1;
      Osint'Elab_Spec;
      Fmap'Elab_Body;
      E224 := E224 + 1;
      E228 := E228 + 1;
      E230 := E230 + 1;
      Osint'Elab_Body;
      E222 := E222 + 1;
      E218 := E218 + 1;
      Osint.M'Elab_Body;
      E334 := E334 + 1;
      Snames'Elab_Body;
      E233 := E233 + 1;
      ALI'ELAB_SPEC;
      ALI'ELAB_BODY;
      E187 := E187 + 1;
      E267 := E267 + 1;
      Sinfo.Nodes'Elab_Spec;
      Einfo.Entities'Elab_Spec;
      Atree'Elab_Spec;
      E249 := E249 + 1;
      Nlists'Elab_Body;
      E255 := E255 + 1;
      Sinfo.Utils'Elab_Spec;
      Aspects'Elab_Body;
      E241 := E241 + 1;
      Atree'Elab_Body;
      E239 := E239 + 1;
      E251 := E251 + 1;
      E258 := E258 + 1;
      Sinput'Elab_Spec;
      E263 := E263 + 1;
      E260 := E260 + 1;
      Sem_Aux'Elab_Spec;
      E284 := E284 + 1;
      E296 := E296 + 1;
      Stringt'Elab_Body;
      E276 := E276 + 1;
      Erroutc'Elab_Spec;
      E274 := E274 + 1;
      Lib'Elab_Spec;
      E282 := E282 + 1;
      E280 := E280 + 1;
      Errout'Elab_Spec;
      Errout'Elab_Body;
      E272 := E272 + 1;
      Fname.Uf'Elab_Body;
      E300 := E300 + 1;
      E298 := E298 + 1;
      E294 := E294 + 1;
      E325 := E325 + 1;
      Make_Util'Elab_Body;
      E323 := E323 + 1;
      E288 := E288 + 1;
      ALI.UTIL'ELAB_SPEC;
      ALI.UTIL'ELAB_BODY;
      E237 := E237 + 1;
      E330 := E330 + 1;
      Switch.M'Elab_Body;
      E336 := E336 + 1;
      Tempdir'Elab_Body;
      E340 := E340 + 1;
      E332 := E332 + 1;
      E328 := E328 + 1;
      Make'Elab_Body;
      E105 := E105 + 1;
      E002 := E002 + 1;
   end adainit;

   procedure Ada_Main_Program;
   pragma Import (Ada, Ada_Main_Program, "_ada_gnatmake");

   function main
     (argc : Integer;
      argv : System.Address;
      envp : System.Address)
      return Integer
   is
      procedure Initialize (Addr : System.Address);
      pragma Import (C, Initialize, "__gnat_initialize");

      procedure Finalize;
      pragma Import (C, Finalize, "__gnat_finalize");
      SEH : aliased array (1 .. 2) of Integer;

      Ensure_Reference : aliased System.Address := Ada_Main_Program_Name'Address;
      pragma Volatile (Ensure_Reference);

   begin
      if gnat_argc = 0 then
         gnat_argc := argc;
         gnat_argv := argv;
      end if;
      gnat_envp := envp;

      Initialize (SEH'Address);
      adainit;
      Ada_Main_Program;
      adafinal;
      Finalize;
      return (gnat_exit_status);
   end;

--  BEGIN Object file/option list
   --   ../rts/ada.o
   --   ../rts/a-charac.o
   --   ../rts/a-chlat1.o
   --   ../rts/interfac.o
   --   ../rts/system.o
   --   ../rts/s-addope.o
   --   ../rts/s-atoope.o
   --   ../rts/s-io.o
   --   ../rts/s-parame.o
   --   ../rts/s-crtl.o
   --   ../rts/i-cstrea.o
   --   ../rts/s-spark.o
   --   ../rts/s-spcuop.o
   --   ../rts/s-stoele.o
   --   ../rts/s-retsta.o
   --   ../rts/s-stache.o
   --   ../rts/s-strhas.o
   --   ../rts/s-htable.o
   --   ../rts/s-string.o
   --   ../rts/s-traent.o
   --   ../rts/s-unstyp.o
   --   ../rts/s-utf_32.o
   --   ../rts/s-wchcon.o
   --   ../rts/s-wchjis.o
   --   ../rts/s-wchcnv.o
   --   ../rts/s-carun8.o
   --   ../rts/s-conca2.o
   --   ../rts/s-crc32.o
   --   ../rts/s-traceb.o
   --   ../rts/s-excdeb.o
   --   ../rts/s-valllu.o
   --   ../rts/s-vallli.o
   --   ../rts/s-valuns.o
   --   ../rts/s-valint.o
   --   ../rts/s-valuti.o
   --   ../rts/s-wchstw.o
   --   ../rts/a-elchha.o
   --   ../rts/a-exctra.o
   --   ../rts/s-addima.o
   --   ../rts/s-bitops.o
   --   ../rts/s-boustr.o
   --   ../rts/s-casuti.o
   --   ../rts/s-exctab.o
   --   ../rts/a-contai.o
   --   ../rts/a-ioexce.o
   --   ../rts/a-numeri.o
   --   ../rts/a-nubinu.o
   --   ../rts/a-string.o
   --   ../rts/a-strmap.o
   --   ../rts/a-stmaco.o
   --   ../rts/i-c.o
   --   ../rts/s-atopri.o
   --   ../rts/s-except.o
   --   ../rts/s-excmac.o
   --   ../rts/s-win32.o
   --   ../rts/a-chahan.o
   --   ../rts/s-aotase.o
   --   ../rts/s-exctra.o
   --   ../rts/s-memory.o
   --   ../rts/s-mmosin.o
   --   ../rts/s-mmap.o
   --   ../rts/s-objrea.o
   --   ../rts/s-os_lib.o
   --   ../rts/s-secsta.o
   --   ../rts/s-soliin.o
   --   ../rts/s-soflin.o
   --   ../rts/s-stalib.o
   --   ../rts/s-trasym.o
   --   ../rts/s-widuns.o
   --   ../rts/s-imgint.o
   --   ../rts/a-except.o
   --   ../rts/s-imguns.o
   --   ../rts/s-dwalin.o
   --   ../rts/a-assert.o
   --   ../rts/a-comlin.o
   --   ../rts/a-strsea.o
   --   ../rts/a-strfix.o
   --   ../rts/a-stuten.o
   --   ../rts/a-suenst.o
   --   ../rts/a-suewst.o
   --   ../rts/a-suezst.o
   --   ../rts/a-tags.o
   --   ../rts/a-sttebu.o
   --   ../rts/a-stbuut.o
   --   ../rts/gnat.o
   --   ../rts/g-casuti.o
   --   ../rts/g-htable.o
   --   ../rts/g-os_lib.o
   --   ../rts/g-string.o
   --   ../rts/s-arit64.o
   --   ../rts/s-atocou.o
   --   ../rts/s-fatflt.o
   --   ../rts/s-fatlfl.o
   --   ../rts/s-fatllf.o
   --   ../rts/s-multip.o
   --   ../rts/s-oscons.o
   --   ../rts/s-putima.o
   --   ../rts/a-stream.o
   --   ../rts/s-ficobl.o
   --   ../rts/s-finroo.o
   --   ../rts/a-finali.o
   --   ../rts/a-conhel.o
   --   ../rts/s-fileio.o
   --   ../rts/s-stopoo.o
   --   ../rts/s-finmas.o
   --   ../rts/s-spsufi.o
   --   ../rts/s-stposu.o
   --   ../rts/s-statxd.o
   --   ../rts/s-stratt.o
   --   ../rts/a-strunb.o
   --   ../rts/s-tasloc.o
   --   ../rts/s-winext.o
   --   ../rts/s-osprim.o
   --   ../rts/a-calend.o
   --   ../rts/a-textio.o
   --   ../rts/g-byorma.o
   --   ../rts/g-ctrl_c.o
   --   ../rts/g-dirope.o
   --   ../rts/g-dynhta.o
   --   ../rts/s-assert.o
   --   ../rts/s-filatt.o
   --   ../rts/s-pooglo.o
   --   ../rts/s-regexp.o
   --   ../rts/a-dirval.o
   --   ../rts/a-direct.o
   --   ../rts/a-dhfina.o
   --   ../rts/g-regexp.o
   --   ../rts/g-comlin.o
   --   ./debug.o
   --   ./types.o
   --   ./alloc.o
   --   ./gnatvsn.o
   --   ./hostparm.o
   --   ./krunch.o
   --   ./csets.o
   --   ./opt.o
   --   ./output.o
   --   ./rident.o
   --   ./seinfo.o
   --   ./sfn_scan.o
   --   ./stand.o
   --   ./stylesw.o
   --   ./table.o
   --   ./elists.o
   --   ./uintp.o
   --   ./urealp.o
   --   ./warnsw.o
   --   ./widechar.o
   --   ./namet.o
   --   ./butil.o
   --   ./binderr.o
   --   ./casing.o
   --   ./err_vars.o
   --   ./fmap.o
   --   ./sdefault.o
   --   ./targparm.o
   --   ./osint.o
   --   ./fname.o
   --   ./osint-m.o
   --   ./sinfo.o
   --   ./snames.o
   --   ./ali.o
   --   ./einfo.o
   --   ./scans.o
   --   ./einfo-entities.o
   --   ./nlists.o
   --   ./aspects.o
   --   ./atree.o
   --   ./einfo-utils.o
   --   ./sinfo-nodes.o
   --   ./sinput.o
   --   ./sinfo-utils.o
   --   ./sem_aux.o
   --   ./sinput-c.o
   --   ./stringt.o
   --   ./erroutc.o
   --   ./uname.o
   --   ./lib.o
   --   ./errout.o
   --   ./fname-uf.o
   --   ./fname-sf.o
   --   ./styleg.o
   --   ./errutil.o
   --   ./make_util.o
   --   ./scng.o
   --   ./ali-util.o
   --   ./switch.o
   --   ./switch-m.o
   --   ./tempdir.o
   --   ./usage.o
   --   ./makeusg.o
   --   ./make.o
   --   ./gnatmake.o
   --   -L../rts/
   --   -L./
   --   -L/mnt/share/packages/gccmake/gcc-13/gcc/gcc.x86_64/src/gcc-13.4.0/gcc/ada/
   --   -L/usr/lib/gcc/x86_64-pc-cygwin/13/adalib/
   --   -static
   --   -lgnat
   --   -Wl,--stack=0x2000000
--  END Object file/option list   

end ada_main;
