pragma Style_Checks (Off);
with Osint; use Osint;
package body Sdefault is
   S0 : constant String := "/usr/";
   S1 : constant String := "/usr/lib/gcc/x86_64-pc-cygwin/13/adainclude/";
   S2 : constant String := "/usr/lib/gcc/x86_64-pc-cygwin/13/adalib/";
   S3 : constant String := "x86_64-pc-cygwin/";
   S4 : constant String := "/usr/lib/gcc/x86_64-pc-cygwin/13/";
   function Include_Dir_Default_Name return String_Ptr is
   begin
      return Relocate_Path (S0, S1);
   end Include_Dir_Default_Name;
   function Object_Dir_Default_Name return String_Ptr is
   begin
      return Relocate_Path (S0, S2);
   end Object_Dir_Default_Name;
   function Target_Name return String_Ptr is
   begin
      return new String'(S3);
   end Target_Name;
   function Search_Dir_Prefix return String_Ptr is
   begin
      return Relocate_Path (S0, S4);
   end Search_Dir_Prefix;
end Sdefault;
